/*-------------------------------------------------------------------------
	Copyright (C) 2008-20013 American Megatrends Inc.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
-------------------------------------------------------------------------*/
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/wait.h>
#include <asm/io.h>
#include <asm/irq.h>
#include "i2c-data.h"
#include "i2c-hardware.h"
#ifdef ARBITRATION_SUPPORT
	#include "i2c-arbitration.h"
#endif

//#define DEBUG 
//#define DEBUG_VERBOSE 

/* Time in ms each I2C byte has to complete */ 
#define I2C_PACKET_TIMEOUT_PER_BYTE       40
#define I2C_PACKET_MAX_TIMEOUT_IN_MS      2500 /* in milliseconds */

/* Prototypes */
int vsc_master_xfer(struct i2c_adapter *i2c_adap, struct i2c_msg *msgs,	int num);
int vsc_slave_xfer(struct i2c_adapter *i2c_adap, char *buf, int	num);
int vsc_xmit_i2c_packet(struct i2c_adapter *i2c_adap);



static u32 
i2c_vsc_wait_event_timeout(int bus, int ms_timeout)
{
	long ret;
	int ms_timeout_final = ms_timeout;

	schedule_timeout_interruptible(1); /* be friendly to the linux kernel (especially the kthread that pets timer4) */

	/* Initialize operation status value */
	vsc_data_ptr[bus].op_status = OP_STATUS_WAIT;

#ifdef DEBUG_VERBOSE
	printk(KERN_INFO "wait_event_timeout() ENTER: bus=%d slave=0x%02x jiffies=%lu\n", bus, (u8)vsc_data_ptr[bus].slave_addr<<1, jiffies);
#endif

#ifdef RESTART_SHORTTIMER
	if(bus == I2C_BUS_2) {
		RESTART_SHORT_TIMER(RESTART_TIMER_VALUE);  /* Restart support fpga short timer */	
	}
#endif
	/* Cap timeout value here since timeout value is based on number of bytes to read/write.  
	We do not want to get into a position where we are waiting too long when devices are failing */
	if (ms_timeout_final > I2C_PACKET_MAX_TIMEOUT_IN_MS) {
		ms_timeout_final = I2C_PACKET_MAX_TIMEOUT_IN_MS;
	}

	/* Clear stop detected status */
	i2c_vsc_write_reg((UINT32)0x40, (UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_BIT_STATUS);

	/* Enable I2C interface by issuing START CONDITION (this fires off interrupt processing) */
	i2c_vsc_write_reg((UINT32)(I2C_CLEAR_ALL_STATUS 
		| I2C_ENABLE_INTERFACE
		| I2C_ENABLE_INTERRUPT
		| I2C_SEND_START
		| I2C_ACKNOWLEDGEMENT),
		(UINT32)i2c_vsc_io_base(bus) + VSC_I2CX_CNTRL);

	/* Wait for interrupt processing to complete */
	ret = wait_event_interruptible_timeout(vsc_data_ptr[bus].vsc_wait, 
		vsc_data_ptr[bus].op_status, msecs_to_jiffies(ms_timeout_final));

#ifdef DEBUG
	/* print only for rare errors to avoid log spam */
	if (vsc_data_ptr[bus].op_status != OP_STATUS_SUCCESS) {
		if ((vsc_data_ptr[bus].scan == 0) ||   
			((vsc_data_ptr[bus].scan == 1) && (vsc_data_ptr[bus].op_status != OP_STATUS_NACK))) {
				printk(KERN_INFO "wait_event_timeout(): bus=%d slave=0x%02x jiffies=%lu, jiffies remaining=%lu, op_status=%s, g_mux_chan=0x%02x\n", bus, vsc_data_ptr[bus].slave_addr<<1, jiffies, (unsigned long)ret, str_op_status[vsc_data_ptr[bus].op_status], g_mux_chan[bus]);
		}
	}
#endif 
	/* if timeout or signal interrupted us (incase wait_event_interruptible_timeout is called), 
	set op_status appropriately */
	if ((ret <= 0) || (OP_STATUS_WAIT == vsc_data_ptr[bus].op_status)) {
		vsc_data_ptr[bus].op_status = OP_STATUS_TIMEOUT;
	}

#ifdef RESTART_SHORTTIMER
	if (bus == I2C_BUS_2) { 
		RESTART_SHORT_TIMER(RESTART_TIMER_VALUE);  /* Restart support fpga short timer */	
	}
#endif

#ifdef DEBUG_VERBOSE
	printk(KERN_INFO "wait_event_timeout() EXIT: bus=%d slave=0x%02x jiffies=%lu\n", bus, (u8)vsc_data_ptr[bus].slave_addr<<1, jiffies);
#endif

	return vsc_data_ptr[bus].op_status;
}


static int 
vsc_send_bytes(struct i2c_adapter *i2c_adap, struct i2c_msg msg)
{
	int i, retval;
	u8 dev_stat_index;

	/* Set flag to Master Transfer Mode */
	vsc_data_ptr[i2c_adap->nr].state = STATE_MASTER_XMIT;
	
	/* Copy write data to buffer */
	for (i=0; i<msg.len; i++) {
		vsc_data_ptr[i2c_adap->nr].TX_data[i] = msg.buf[i];
	}
   
	/* If no data to write set scan flag (means this is just an i2c scan) */
	vsc_data_ptr[i2c_adap->nr].scan = msg.len ? 0 : 1;

	/* Set transmit/receive lengths and indexes */
	vsc_data_ptr[i2c_adap->nr].TX_len = msg.len;
	vsc_data_ptr[i2c_adap->nr].TX_index = 0;
 	vsc_data_ptr[i2c_adap->nr].MasterRX_index = 0;
	vsc_data_ptr[i2c_adap->nr].MasterRX_len = 0;
   
	/* Transmit I2C packet, if transmission was successful */
	retval = vsc_xmit_i2c_packet(i2c_adap);

	/* If transmission was successful */
	if (!retval) {

		/* Get index to device statistics structure */
		dev_stat_index = i2c_vsc_get_dev_stat_index(i2c_adap->nr, vsc_data_ptr[i2c_adap->nr].slave_addr, g_mux_chan[i2c_adap->nr]);

		/* If device stats available for this device */
		if (i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].present == VALID) {

			/* Update bus stats (total xmits) */
			i2c_bus_status_ptr[i2c_adap->nr].TX_index += vsc_data_ptr[i2c_adap->nr].TX_index;

			/* Update device stats (total device xmits) */
			i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].xmit_cnt += vsc_data_ptr[i2c_adap->nr].TX_index;
		}
	}

	return retval; 
}


static int 
vsc_receive_bytes(struct i2c_adapter *i2c_adap, struct i2c_msg msg)
{
	int i, retval;
	u8 dev_stat_index;

	/* Set flag to Master Receive Mode */
	vsc_data_ptr[i2c_adap->nr].state = STATE_MASTER_RECV;

	/* Set transmit/receive lengths and indexes */
	vsc_data_ptr[i2c_adap->nr].MasterRX_index = 0;
	vsc_data_ptr[i2c_adap->nr].MasterRX_len = msg.len;
	vsc_data_ptr[i2c_adap->nr].TX_len = 0;
	vsc_data_ptr[i2c_adap->nr].TX_index = 0;

	/* If no data to read set scan flag (means this is just an i2c scan) */
	vsc_data_ptr[i2c_adap->nr].scan = msg.len ? 0 : 1;

	/* Transmit I2C packet, if transmission was successful */
	retval = vsc_xmit_i2c_packet(i2c_adap);

	/* If transmission was successful */
	if (!retval) {

		/* Get index to device statistics structure */
		dev_stat_index = i2c_vsc_get_dev_stat_index(i2c_adap->nr, vsc_data_ptr[i2c_adap->nr].slave_addr, g_mux_chan[i2c_adap->nr]);

		/* If device stats available for this device */
		if (i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].present == VALID) {

			/* Update receive data statistics */
			i2c_bus_status_ptr[i2c_adap->nr].MasterRX_index  += vsc_data_ptr[i2c_adap->nr].MasterRX_index;
			i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].recv_cnt += vsc_data_ptr[i2c_adap->nr].MasterRX_index;
		}

		/* Get response data from the buffer */
		for (i=0; i<msg.len; i++)
			msg.buf[i] = vsc_data_ptr[i2c_adap->nr].MasterRX_data[i];
	}

	return retval;
}


static int 
vsc_send_receive_bytes(struct i2c_adapter *i2c_adap, struct i2c_msg write_msg, struct i2c_msg read_msg)
{
	int i, retval;
	unsigned char dev_stat_index;

	/* Set flag to Master Transfer Mode */
	vsc_data_ptr[i2c_adap->nr].state = STATE_MASTER_XMIT_RECV;

	/* Copy write data to buffer */
	for (i=0; i<write_msg.len; i++) {
		vsc_data_ptr[i2c_adap->nr].TX_data[i] = write_msg.buf[i];
	}

	/* Set transmit/receive lengths and indexes */
	vsc_data_ptr[i2c_adap->nr].TX_len = write_msg.len;
	vsc_data_ptr[i2c_adap->nr].TX_index = 0;
	vsc_data_ptr[i2c_adap->nr].MasterRX_index = 0;
	vsc_data_ptr[i2c_adap->nr].MasterRX_len = read_msg.len;

 	/* No scan (scan only performed with WRITE flag) */
	vsc_data_ptr[i2c_adap->nr].scan = 0;
	
	/* If no data to read and write return error */
	if ((write_msg.len == 0) && (read_msg.len == 0)) {
		return -(int)OP_STATUS_UNKNOWN;
	}

	/* Transmit I2C packet, if transmission was successful */
	retval = vsc_xmit_i2c_packet(i2c_adap);
 
	/* If transmission was successful */
	if (!retval) {

		/* Get index to device statistics structure */
		dev_stat_index = i2c_vsc_get_dev_stat_index(i2c_adap->nr, vsc_data_ptr[i2c_adap->nr].slave_addr, g_mux_chan[i2c_adap->nr]);

		/* If device stats available for this device */
		if (i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].present == VALID) {

			/* Update bus stats (byte transmitted count and byte received count) */
			i2c_bus_status_ptr[i2c_adap->nr].TX_index  += vsc_data_ptr[i2c_adap->nr].TX_index;
			i2c_bus_status_ptr[i2c_adap->nr].MasterRX_index  += vsc_data_ptr[i2c_adap->nr].MasterRX_index;

			/* Update device stats (byte transmitted count and byte received count) */
			i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].xmit_cnt += vsc_data_ptr[i2c_adap->nr].TX_index;
			i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].recv_cnt += vsc_data_ptr[i2c_adap->nr].MasterRX_index;
		}

		/* Copy read data to buffer */
		for (i=0; i<read_msg.len; i++) 
			read_msg.buf[i] = vsc_data_ptr[i2c_adap->nr].MasterRX_data[i];

	}

	return retval;
}


static int 
send_slave_addr(struct i2c_adapter *i2c_adap, int addr, int flags)
{
	/* Save current slave address to global variable */
	vsc_data_ptr[i2c_adap->nr].slave_addr = addr;

	/* Set slave address and read/write bit in data register (sent after start condition) */
	if (flags & I2C_M_RD) {
		/* Read - append R/W bit as 1(one) */
		i2c_vsc_write_reg((addr << 1) | READ_OPERATION_BIT, 
			(UINT32)i2c_vsc_io_base(i2c_adap->nr) + VSC_I2CX_DATA);
	} else {
		/* Write - append R/W bit as 0(zero) */
		i2c_vsc_write_reg((addr << 1) | WRITE_OPERATION_BIT, 
			(UINT32)i2c_vsc_io_base(i2c_adap->nr) + VSC_I2CX_DATA);
	}
	
	return 0;
}


int 
vsc_master_xfer(struct i2c_adapter *i2c_adap, struct i2c_msg *msgs, int num)
{
	int i = 0;
	int retval = 0;
	unsigned char dev_stat_index;
        u8 val_mlbr;

#ifdef DEBUG_VERBOSE
	printk(KERN_ERR "vsc_master_xfer: bus=%d ", i2c_adap->nr);
	for (i=0; i<num-1; i++)
		printk("msgs[%d].addr=%02X:%d, ", i, msgs[i].addr<<1, msgs[i].segnum);
	printk("msgs[%d].addr=%02X:%d\n", num-1, msgs[num-1].addr<<1, msgs[num-1].segnum);
#endif

	/* dev_stat_index is obtained in other functions that do not know the slave address.  So we
	must save the slave address here for those functions (for example: vsc_i2c_bus_recovery())*/
	vsc_data_ptr[i2c_adap->nr].slave_addr = msgs[0].addr;
	g_mux_chan[i2c_adap->nr] = msgs[0].segnum;

#ifdef ARBITRATION_SUPPORT
	if (i2c_adap->nr == I2C_BUS_2) {

		/* if bus request successful */
		if (i2c_request_shared_bus(g_mux_chan[i2c_adap->nr]) != STATUS_SUCCESS) {
			return -(int)OP_STATUS_BUSY;
		}

		/* wait for bus grant and if bus granted */
		if (i2c_grant_shared_bus(g_mux_chan[i2c_adap->nr]) != STATUS_SUCCESS) {
			return -(int)OP_STATUS_BUSY;
		}
		
		/* (re)start short time */
		supfpga_mm_reg8_read(SUPFPGA_MLBR_REG, &val_mlbr);
		fpga_start_short_timer(val_mlbr);		
	}
#endif
			
	dev_stat_index = i2c_vsc_get_dev_stat_index(i2c_adap->nr, vsc_data_ptr[i2c_adap->nr].slave_addr, g_mux_chan[i2c_adap->nr]);

	/* If too many consecutive bus busy failures have occured on this bus, it might be hung */
	if (i2c_bus_status_ptr[i2c_adap->nr].bus_conseq_busy_cnt >= MAX_BUS_BUSY_CNT) {
#ifdef DEBUG
		printk(KERN_ERR "I2C%d: Bus consequtively busy %d times (vsc_master_xfer) : bus_busy_cnt=%d bus_conseq_busy_cnt=%d\n", i2c_adap->nr, 
			MAX_BUS_BUSY_CNT, (int)i2c_bus_status_ptr[i2c_adap->nr].bus_busy_cnt, (int)i2c_bus_status_ptr[i2c_adap->nr].bus_conseq_busy_cnt);
#endif
		i2c_bus_status_ptr[i2c_adap->nr].bus_conseq_busy_cnt = 0;

		/* Is the I2C bus hung? */
		if (vsc_is_i2c_bus_hung(i2c_adap)) {

			/* Perform I2C bus recovery */
			vsc_i2c_bus_recovery(i2c_adap);

			/* Reinitialise I2C internal data and hardware */
			i2c_init_internal_data(i2c_adap->nr);
			i2c_init_hardware(i2c_adap->nr);

#ifdef ARBITRATION_SUPPORT
			if (i2c_adap->nr == I2C_BUS_2) {
				/* release shared bus */
				i2c_release_shared_bus(g_mux_chan[i2c_adap->nr]);
			}
#endif			
			return -(int)OP_STATUS_BUSY;
		}
	}
   
	/* Check if bus busy */
	if (i2c_wait_for_bus_free(i2c_adap->nr)) {

		/* Update bus busy stats */
		i2c_bus_status_ptr[i2c_adap->nr].bus_busy_cnt++;
		i2c_bus_status_ptr[i2c_adap->nr].bus_conseq_busy_cnt++;

#ifdef DEBUG
		printk(KERN_WARNING "I2C%d: Bus busy (vsc_master_xfer) : bus_busy_cnt=%d bus_conseq_busy_cnt=%d\n", i2c_adap->nr, 
			(int)i2c_bus_status_ptr[i2c_adap->nr].bus_busy_cnt, (int)i2c_bus_status_ptr[i2c_adap->nr].bus_conseq_busy_cnt);
#endif
		/* Reinitialise I2C internal data and hardware */
		i2c_init_internal_data(i2c_adap->nr);
		i2c_init_hardware(i2c_adap->nr);

#ifdef ARBITRATION_SUPPORT
		if (i2c_adap->nr == I2C_BUS_2) {
			/* release shared bus */
			i2c_release_shared_bus(g_mux_chan[i2c_adap->nr]);
		}
#endif
		return -(int)OP_STATUS_BUSY;
	}

	/* Bus is free so reset bus consequtive busy counter */
	i2c_bus_status_ptr[i2c_adap->nr].bus_conseq_busy_cnt = 0;
 
	/* for each I2C packet */
	for (i=0; ((retval>=0) && (i<num)); i++) {

		/* Reinitialise I2C internal data and hardware */
		i2c_init_internal_data(i2c_adap->nr);
		i2c_init_hardware(i2c_adap->nr);
		udelay(10);

		/* Disallow access to our own slave address */
		if ((u8)msgs[i].addr == (u8)(slave_addr[i2c_adap->nr])) {
#ifdef ARBITRATION_SUPPORT
			if (i2c_adap->nr == I2C_BUS_2) {
				/* release shared bus */
				i2c_release_shared_bus(g_mux_chan[i2c_adap->nr]);
			}
#endif
			return -(int)OP_STATUS_UNKNOWN;
		}

		/* Setup slave address */
		if (msgs[i].flags & I2C_M_RD) {
			send_slave_addr(i2c_adap, msgs[i].addr, READ_OPERATION_BIT);
		} else {
			send_slave_addr(i2c_adap, msgs[i].addr, WRITE_OPERATION_BIT);
		} 

		/* Update g_mux_chan and dev_stat_index for this packet */
		g_mux_chan[i2c_adap->nr] = msgs[i].segnum;
		dev_stat_index = i2c_vsc_get_dev_stat_index(i2c_adap->nr, vsc_data_ptr[i2c_adap->nr].slave_addr, g_mux_chan[i2c_adap->nr]);

		/* If this msg is a write operation */
		if ((msgs[i].flags & I2C_M_RD) == 0) {
			/* If another operation follows this write operation */
			if ((i + 1) < num) {
				/* If the next operation is a read operation to same slave address */
				if ((msgs[i+1].flags & I2C_M_RD) && (msgs[i+1].addr == msgs[i].addr)) {
					/* Perform write operation followed by repeated start followed by read operation */
					retval = vsc_send_receive_bytes(i2c_adap, msgs[i], msgs[i+1]);
					/* Increment i one additional time since we just performed two operations */
					i++;
				} else {
					/* Otherwise perform a single master write operation */
					retval = vsc_send_bytes(i2c_adap, msgs[i]);
				}
			} else {
				/* Otherwise perform a single master write operation */
				retval = vsc_send_bytes(i2c_adap, msgs[i]);
			}
		/* If this is a read operation */
		} else {
			/* Perform a single master read operation */
			retval = vsc_receive_bytes(i2c_adap, msgs[i]);
		} 

		/* If operation failed */
		if (0 != retval) {

			/* If this was not a scan or if this was non nack error from i2c scan */
			if ((vsc_data_ptr[i2c_adap->nr].scan == 0) || ((vsc_data_ptr[i2c_adap->nr].scan == 1) && (retval != -(int)OP_STATUS_NACK))) {

				/* If device stats available for this device */
				if (i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].present == VALID) {

					/* If accesses to this device have repeatedly failed (conseq_fail_cnt updated from vsc_xmit_i2c_packet())*/
					if (i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].conseq_fail_cnt >= MAX_FAILURES_PER_DEVICE) {
#ifdef DEBUG
						printk(KERN_ERR "I2C%d:0x%02x: Device consequtive failure count exceeded (vsc_master_xfer) : conseq_fail_cnt=%d bus_conseq_busy_cnt=%d conseq_busy_cnt=%d\n", 
							i2c_adap->nr, msgs[i].addr<<1, 
							(int)i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].conseq_fail_cnt,
							(int)i2c_bus_status_ptr[i2c_adap->nr].bus_conseq_busy_cnt,
							(int)i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].conseq_busy_cnt);
#endif
						/* Check if bus hung? */
						if (vsc_is_i2c_bus_hung(i2c_adap)) {

							/* If so perform bus recovery */
							vsc_i2c_bus_recovery(i2c_adap);
						}

						/* Reset consequtive device failure count */
						i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].conseq_fail_cnt = 0;
					}

					/* If we repeatedly detect bus busy when accessing this device (conseq_busy_cnt updated from vsc_xmit_i2c_packet()) */
					if (i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].conseq_busy_cnt >= MAX_BUS_BUSY_CNT) {
#ifdef DEBUG
						printk(KERN_ERR "I2C%d:0x%02x: Device consequtive busy count exceeded (vsc_master_xfer) : bus_conseq_busy_cnt=%d conseq_busy_cnt=%d conseq_fail_cnt=%d\n", 
							i2c_adap->nr, msgs[i].addr<<1, 
							(int)i2c_bus_status_ptr[i2c_adap->nr].bus_conseq_busy_cnt,
							(int)i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].conseq_busy_cnt,
							(int)i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].conseq_fail_cnt);
#endif
						/* Check if bus hung? */
						if (vsc_is_i2c_bus_hung(i2c_adap)) {

							/* If so perform bus recovery */
							vsc_i2c_bus_recovery(i2c_adap);
						}

						/* Reset consequtive device busy count */
						i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].conseq_busy_cnt = 0;
					}
				}
			}

			/* Reinitialise I2C internal data and hardware */
			i2c_init_internal_data(i2c_adap->nr);
			i2c_init_hardware(i2c_adap->nr);

#ifdef ARBITRATION_SUPPORT
			if (i2c_adap->nr == I2C_BUS_2) {
				/* release shared bus */
				i2c_release_shared_bus(g_mux_chan[i2c_adap->nr]);
			}
#endif
	
			return retval;
		}
	}

	/* Reinitialise I2C internal data and hardware */
	i2c_init_internal_data(i2c_adap->nr);
	i2c_init_hardware(i2c_adap->nr);

#ifdef ARBITRATION_SUPPORT
	if (i2c_adap->nr == I2C_BUS_2) {
		/* release shared bus */
		i2c_release_shared_bus(g_mux_chan[i2c_adap->nr]);
	}
#endif

	/* Return number of i2c packets transmitted */
	return num;
}


int 
vsc_slave_xfer(struct i2c_adapter *i2c_adap,
			char *buf, int num)
{
	/* Currently not implemented */
	return 0;
}


int 
vsc_xmit_i2c_packet(struct i2c_adapter *i2c_adap)
{
	u32 op_status;
	u8 dev_stat_index;
#ifdef DEBUG
	char str_xmit_type[32];
#endif
	int timeout;
	int retval=-(int)OP_STATUS_UNKNOWN;
	int state = vsc_data_ptr[i2c_adap->nr].state; 

#ifdef DEBUG	
	if (state == STATE_MASTER_XMIT)
		strcpy(str_xmit_type, "vsc_send_bytes");
	else if (state == STATE_MASTER_RECV)
		strcpy(str_xmit_type, "vsc_receive_bytes");
	else if (state == STATE_MASTER_XMIT_RECV)
		strcpy(str_xmit_type, "vsc_send_receive_bytes");
	else
		return -(int)OP_STATUS_UNKNOWN;
#else
	if ((state != STATE_MASTER_XMIT) && (state != STATE_MASTER_RECV) && (state != STATE_MASTER_XMIT_RECV)) {
		return -(int)OP_STATUS_UNKNOWN;
	}
#endif

	/* Get index to device statistics structure */
	dev_stat_index = i2c_vsc_get_dev_stat_index(i2c_adap->nr, vsc_data_ptr[i2c_adap->nr].slave_addr, g_mux_chan[i2c_adap->nr]);

	/* Update statistics (mode) */
	if (i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].present == VALID) {
		i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].mode = state;
	}

	/* Make sure bus is still free before taking ownership (i.e. before issuing start condition) */
	if (i2c_wait_for_bus_free(i2c_adap->nr)) {
#ifdef DEBUG
		printk(KERN_ERR "I2C%d:0x%02x: Bus busy (%s) : bus_conseq_busy_cnt=%d conseq_busy_cnt=%d conseq_fail_cnt=%d nack_cnt=%d\n", 
			i2c_adap->nr, vsc_data_ptr[i2c_adap->nr].slave_addr<<1, str_xmit_type,
			(int)i2c_bus_status_ptr[i2c_adap->nr].bus_conseq_busy_cnt,
			(int)i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].conseq_busy_cnt,
			(int)i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].conseq_fail_cnt,
			(int)i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].nack_cnt);
#endif

		/* Update bus stats */
		i2c_bus_status_ptr[i2c_adap->nr].bus_busy_cnt++;
		i2c_bus_status_ptr[i2c_adap->nr].bus_conseq_busy_cnt++;

		/* Update device stats (busy count) */
		i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].busy_cnt++;
		i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].conseq_busy_cnt++;

		retval = -((int)OP_STATUS_BUSY);
	}
	else {

		/* Bus free (even potentially behind mux) so reset consequtive device busy busy count */
		if (i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].present == VALID) {
			i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].conseq_busy_cnt = 0;
		}

		/* Issue start condition and wait here for transmission to complete (rest of the logic handled inside interrupt handler) */
		timeout = (((vsc_data_ptr[i2c_adap->nr].TX_len + vsc_data_ptr[i2c_adap->nr].MasterRX_len) + 1) * I2C_PACKET_TIMEOUT_PER_BYTE); 
		op_status = i2c_vsc_wait_event_timeout(i2c_adap->nr, timeout); 

		/* If packet transmission was successful */
		if (OP_STATUS_SUCCESS == op_status) {

			/* Reset consequtive device failure count and busy count after successfully transmission */
			if (i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].present == VALID) {
				i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].conseq_fail_cnt = 0;
			}

			retval = 0; /* Return success */
		}
		else { /* If transmission was unsuccessful */

			/* If NACK error from scan avoid spamming syslog and unnecessarily invoking recovery from scan NACK errors */
			if ((vsc_data_ptr[i2c_adap->nr].scan) && (OP_STATUS_NACK == op_status)) {
					retval = -(int)OP_STATUS_NACK; /* Return NACK */
			}
			else {
				/* Track bus failures */
				i2c_bus_status_ptr[i2c_adap->nr].bus_fail_cnt++;

				/* Track of failures per device */
				if (i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].present == VALID) {
					i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].conseq_fail_cnt++;
					i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].failure_cnt++;
				}

				/* If NACK error */
				if (OP_STATUS_NACK == op_status) {

					/* Track NACK errors */
					if (i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].present == VALID) {
						i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].nack_cnt++;
					}
				}
#ifdef DEBUG
				printk(KERN_ERR "I2C%d:0x%02x: Unsuccessful op_status=%s (%s) : conseq_busy_cnt=%d conseq_fail_cnt=%d nack_cnt=%d\n", 
					i2c_adap->nr, vsc_data_ptr[i2c_adap->nr].slave_addr<<1, str_op_status[op_status], str_xmit_type,
					(int)i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].conseq_busy_cnt,
					(int)i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].conseq_fail_cnt,
					(int)i2c_bus_status_ptr[i2c_adap->nr].dev_stat[dev_stat_index].nack_cnt);
#endif
				retval = -(int)op_status; /* Return specific failure */
			}
		}
	}

	return retval; /* return error code (0 success : <0 error) */
}
