/*
 *
 * Copyright (c) 2011 - Network Appliance Inc.
 *
 */
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/cdev.h>
#include <linux/sched.h>
#include <linux/delay.h>

#include <asm/uaccess.h>

#include <tap.h>
#include "tap_gpio.h"

#define TAP_MAX_DEVS 50

typedef union tap_ioctl {
	tap_scan_t scan;
	tap_config_t cfg;
} tap_ioctl_t;

static tap_ioctl_t tap_ioctl_cmd;

/* The state transition table for a TAP controller */
int tap_sm[TAP_NR_STATES][2] = {
  	/* TAP_TEST_LOGIC_RESET */ { TAP_RUN_TEST_IDLE, TAP_TEST_LOGIC_RESET },
  	/* TAP_RUN_TEST_IDLE    */ { TAP_RUN_TEST_IDLE, TAP_SELECT_DR_SCAN   },
  	/* TAP_SELECT_DR_SCAN   */ { TAP_CAPTURE_DR,    TAP_SELECT_IR_SCAN   },
  	/* TAP_CAPTURE_DR       */ { TAP_SHIFT_DR,      TAP_EXIT_1_DR        }, 
  	/* TAP_SHIFT_DR         */ { TAP_SHIFT_DR,      TAP_EXIT_1_DR        },
  	/* TAP_EXIT_1_DR        */ { TAP_PAUSE_DR,      TAP_UPDATE_DR        },
  	/* TAP_PAUSE_DR         */ { TAP_PAUSE_DR,      TAP_EXIT_2_DR        },
  	/* TAP_EXIT_2_DR        */ { TAP_SHIFT_DR,      TAP_UPDATE_DR        },
  	/* TAP_UPDATE_DR        */ { TAP_RUN_TEST_IDLE, TAP_SELECT_DR_SCAN   },
  	/* TAP_SELECT_IR_SCAN   */ { TAP_CAPTURE_IR,    TAP_TEST_LOGIC_RESET },
  	/* TAP_CAPTURE_IR       */ { TAP_SHIFT_IR,      TAP_EXIT_1_IR        },
  	/* TAP_SHIFT_IR         */ { TAP_SHIFT_IR,      TAP_EXIT_1_IR        },
  	/* TAP_EXIT_1_IR        */ { TAP_PAUSE_IR,      TAP_UPDATE_IR        },
  	/* TAP_PAUSE_IR         */ { TAP_PAUSE_IR,      TAP_EXIT_2_IR        },
  	/* TAP_EXIT_2_IR        */ { TAP_SHIFT_IR,      TAP_UPDATE_IR        },
  	/* TAP_UPDATE_IR        */ { TAP_RUN_TEST_IDLE, TAP_SELECT_DR_SCAN   },
};


static int tap_next[TAP_NR_STATES][TAP_NR_STATES] = { };

struct tap_controller {
	tap_config_t cfg;
	int count;
	dev_t dev;
	struct cdev cdev;
	tap_state_t state;
} tap;

module_param_named(tdo, tap.cfg.tdo, int, S_IRUGO);
module_param_named(tdi, tap.cfg.tdi, int, S_IRUGO);
module_param_named(tck, tap.cfg.tck, int, S_IRUGO);
module_param_named(tms, tap.cfg.tms, int, S_IRUGO);
module_param_named(trst, tap.cfg.trst, int, S_IRUGO);


void tap_tdo_set(struct tap_controller *t, int value)
{
	tap_gpio_set_value(t->cfg.tdo, value);
}

int tap_tdi_get(struct tap_controller *t)
{
	int val;
	val = tap_gpio_get_value(t->cfg.tdi);
	return val;
}

static void
tap_tms_send(struct tap_controller *t, int value)
{
	tap_gpio_set_value(t->cfg.tms, value);
	/* burt1035018 for TMS only add an additional 1 micro second delay after setting TMS value to 
	 * allow time for it to settle before pulling clock high.*/
	udelay(1);
	tap_gpio_set_value(t->cfg.tck, 1);
	udelay(1);
	tap_gpio_set_value(t->cfg.tck, 0);
	t->state = tap_sm[t->state][value];
}

static void tap_tms_line_low(struct tap_controller *t)
{
	tap_gpio_set_value(t->cfg.tms, 0);
}

static void tap_tms_line_high(struct tap_controller *t)
{
	tap_gpio_set_value(t->cfg.tms, 1);
}

static void tap_tck_line_low(struct tap_controller *t)
{
	tap_gpio_set_value(t->cfg.tck, 0);
}

static void tap_tck_line_high(struct tap_controller *t)
{
	tap_gpio_set_value(t->cfg.tck, 1);
}


static void tap_trst_line_low(struct tap_controller *t)
{
	tap_gpio_set_value(t->cfg.trst, 0);
}

static void tap_trst_line_high(struct tap_controller *t)
{
	tap_gpio_set_value(t->cfg.trst, 1);
}


static void
tap_compute_shortest_path(void)
{
	int path[TAP_NR_STATES][TAP_NR_STATES];
	int i, j, k;
	int new_dist;

	/* initialize to infinity */
	for (i = 0; i < TAP_NR_STATES; i++) {
		for (j = 0; j < TAP_NR_STATES; j++) {
			path[i][j] = TAP_NR_STATES;
		}
	}

	/* if there is a direct transition then shortest path length is 1 */
	for (i = 0; i < TAP_NR_STATES; i++) {
		path[i][tap_sm[i][0]] = 1;
		path[i][tap_sm[i][1]] = 1;
	}

	/* if source is same as dest, the shortest path length is 0 */
	for (i = 0; i < TAP_NR_STATES; i++) {
		path[i][i] = 0;
	}

	for (k = 0; k < TAP_NR_STATES; k++) {
		/* for each k compute shortest path from i to j, using nodes 0..k */
		for (i = 0; i < TAP_NR_STATES; i++) {
			for (j = 0; j < TAP_NR_STATES; j++) {
				new_dist = path[i][k] + path[k][j];
				if (new_dist < path[i][j]) {
					path[i][j] = new_dist;
				}
			}
		}
	}

	for (i = 0; i < TAP_NR_STATES; i++) {
		for (j = 0; j < TAP_NR_STATES; j++) {
			/* which transition takes us closer 0 or 1 */
			if (path[tap_sm[i][0]][j] < path[tap_sm[i][1]][j]) {
				tap_next[i][j] = 0;
			} else {
				tap_next[i][j] = 1;
			}
		}
	}
}

static void tap_reset_init_driver_state(struct tap_controller *t)
{
	// printk(" tap_reset_init_driver_state in driver called , started state is %d \n", t->state ) ;
	t->state = TAP_TEST_LOGIC_RESET; // any state is ok
	// printk(" tap_reset_init_driver_state in driver called , now state is %d \n", t->state ) ;
}

/*
 * Looking at the TAP state diagram, we can see that a TMS input
 * sequence of 11111, is sufficient to bring the state machine
 * to TEST_LOGIC_RESET state irrespective of the starting state.
 */
static void tap_reset(struct tap_controller *t)
{
	int i = 6;

	t->state = TAP_RUN_TEST_IDLE; // any state is ok
	while (i > 0) {
		tap_tms_send(t, 1);
		i--;
	}
}

/*
 * TAP helper functions
 */
static void tap_goto(struct tap_controller *t, tap_state_t dest)
{
	// assert(t->state != TAP_UNKNOWN)
	while (t->state != dest) {
		tap_tms_send(t, tap_next[t->state][dest]);
	}
}


static void tap_enable(void)
{
	/*
	 * Enables JTAG in a vespa. Make this cleaner later.
         */
	*((volatile uint32_t*) 0xb0200838) = 0;
}


static void
tap_shift_n(struct tap_controller *t, unsigned int input[],
	    unsigned int output[], int n)
{
	int i = 0;
	unsigned int mask = 0x1;
	
	while (n > 0) { /* there are more bits to shift */
		tap_tms_send(t, 0);
		if (input) {
			tap_tdo_set(t, (input[i] & mask) ? 1 : 0);
		}
		if (output) {
			if (tap_tdi_get(t)) {
				output[i] |=  mask;
			} else {
				output[i] &= ~mask;
			}
		}
		
		mask <<= 1;
		if (!mask) {
			i++;
			mask = 0x1;
		}
		n--; /* we just finished shifting one bit */
	}
}		

static void
tap_scan_ir_n(struct tap_controller *t, unsigned int input[], unsigned int output[], int n)
{
	tap_goto(t, TAP_CAPTURE_IR);
	tap_shift_n(t, input, output, n);
	tap_goto(t, TAP_RUN_TEST_IDLE);
}

static void
tap_scan_dr_n(struct tap_controller *t, unsigned int input[], unsigned int output[], int n)
{
	tap_goto(t, TAP_CAPTURE_DR);
	tap_shift_n(t, input, output, n);
	tap_goto(t, TAP_RUN_TEST_IDLE);
}

static int
tap_count(struct tap_controller *t)
{
	int i;
	int found = 0;

	tap_goto(t, TAP_SHIFT_IR);

	tap_gpio_set_value(t->cfg.tdo, 1);

	/* Load the BYPASS instruction in each IR */
	for(i = 0; i < TAP_MAX_DEVS * 32; i++) {
		tap_tms_send(t, 0);
	}

	tap_goto(t, TAP_SHIFT_DR);
	tap_gpio_set_value(t->cfg.tdo, 0);

	/* shift zeroes to all bypass registers */
	for (i = 0; i < TAP_MAX_DEVS; i++) {
		tap_tms_send(t, 0);
	}

	/* start shifting ones and count steps till we get a 1 back */
	tap_gpio_set_value(t->cfg.tdo, 1);
	for (i = 0; i < TAP_MAX_DEVS + 1; i++) {
		tap_tms_send(t, 0);
		if (tap_gpio_get_value(t->cfg.tdi)) {
			found = 1;
			break;
		}
	}

	return found ? i : 0;
}


void tap_config_update(struct tap_controller *tc)
{

}



/*
 * The ioctl handler.
 */
long tap_ioctl( struct file *filp,
	      unsigned int cmd, unsigned long arg)
{
	int num;
	int dir_r = _IOC_DIR(cmd) & _IOC_READ;
	int dir_w = _IOC_DIR(cmd) & _IOC_WRITE;
	unsigned int *input = NULL;
	unsigned int *output = NULL;
	int ret_val = 0 ;


	if (_IOC_TYPE(cmd) != TAP_IOCTL_MAGIC) {
		return -ENOTTY;
	}

	if (_IOC_NR(cmd) > TAP_IOCTL_MAX_NR) {
		return -ENOTTY;
	}

	if (dir_r || dir_w) {
		if (!access_ok(VERIFY_WRITE, (void __user *) arg, 
			       _IOC_SIZE(cmd))) {
			return -EFAULT;
		}

		copy_from_user(&tap_ioctl_cmd, (void __user *) arg, 
			       _IOC_SIZE(cmd));
	}

	if (dir_r) {
		output = &tap_ioctl_cmd.scan.output[0];
	}
	if (dir_w) {
		input = &tap_ioctl_cmd.scan.input[0];
	}

	
	switch (cmd) {
		case TAP_IOCTL_ENABLE:
			tap_enable();
			break ;

		case TAP_IOCTL_RESET:
			tap_reset(&tap);
			break ;

		case TAP_IOCTL_INIT_DRIVER_STATE :
			tap_reset_init_driver_state(&tap);
			break ;

		case TAP_IOCTL_TRST_LOW:
			tap_trst_line_low(&tap);
			break ;

		case TAP_IOCTL_TRST_HIGH:
			tap_trst_line_high(&tap);
			break ;

		case TAP_IOCTL_TCK_LOW:
			tap_tck_line_low(&tap);
			break ;

		case TAP_IOCTL_TCK_HIGH:
			tap_tck_line_high(&tap);
			break ;

		case TAP_IOCTL_TMS_LOW:
			tap_tms_line_low(&tap);
			break ;

		case TAP_IOCTL_TMS_HIGH:
			tap_tms_line_high(&tap);
			break ;

		case TAP_IOCTL_COUNT:
			num = tap_count(&tap);
			ret_val = num ;
			break ;

		case TAP_IOCTL_SET_CFG:
			memcpy(&tap.cfg, &tap_ioctl_cmd.cfg, sizeof(tap.cfg));
			break ;

		case TAP_IOCTL_SCAN_DR_R:
		case TAP_IOCTL_SCAN_DR_W:
		case TAP_IOCTL_SCAN_DR_RW:
			tap_scan_dr_n(&tap, input, output, 
			      tap_ioctl_cmd.scan.scan_length); 
			break;


		case TAP_IOCTL_SCAN_IR_R:
		case TAP_IOCTL_SCAN_IR_W:
		case TAP_IOCTL_SCAN_IR_RW:
			tap_scan_ir_n(&tap, input, output, 
			      tap_ioctl_cmd.scan.scan_length);
			break;

		default :
			break ;
	} // End of switch.
	if (dir_r) {
		copy_to_user((void __user *) arg, &tap_ioctl_cmd.scan, 
			     sizeof(tap_ioctl_cmd.scan));
	}
	return ret_val ;
}


struct file_operations tap_fops = {
	.owner = THIS_MODULE,
	.unlocked_ioctl = tap_ioctl,
};



int tap_init_module(void)
{
	int rc;

	rc = alloc_chrdev_region(&tap.dev, 0, 1, "tap"); 
	if (rc) {
		printk(KERN_ERR "failed getting major number");
		return rc;
	}

	cdev_init(&tap.cdev, &tap_fops);

	rc = cdev_add(&tap.cdev, tap.dev, 1);
	if (rc) {
		printk(KERN_ERR "error %d adding device %d:%d\n", 
		       rc, MAJOR(tap.dev), MINOR(tap.dev));
		return rc;
	}

	printk(KERN_ERR "added device %d:%d\n", MAJOR(tap.dev), 
	       MINOR(tap.dev));


	/*
         * Compute shortest paths one time.
         */
	tap_compute_shortest_path();

	tap.count = 0;
	 
	tap.state = TAP_UNKNOWN;
	return 0;
}

void tap_cleanup_module(void)
{
	cdev_del(&tap.cdev);

	/* cleanup_module is never called if registering failed */
	unregister_chrdev_region(tap.dev, 1);
}

module_init(tap_init_module);
module_exit(tap_cleanup_module);

MODULE_DESCRIPTION("TAP Driver");
MODULE_AUTHOR("NetApp Inc");
MODULE_LICENSE("GPL");
